/* (temporary comment)
 *
 * arch/arm/mach-msm/kc_bcm_signal_control.c
 *
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
 *
 * Author: P.Linh
 * Created Date : 2009/11/12
 * Modified date :
 *    2010/01/15  P.Linh Change function name & internal process for new DHD Release (2009/10/29) 
 *                       kc_bcm_wifi_signal_ctl(on/off) -> bcm_wlan_power_on/bcm_wlan_power_off               
 * Content : Control signal of BCM4325 (WiFi/Bluetooth Chip)
 *           Power ON/OFF, REG_ON, WL_RST..... 
 */

#include <linux/module.h>
#include <linux/delay.h>

#include <mach/hs_io_ctl_a.h>
#include <mach/vreg.h>
#include <mach/kc_bcm_signal_control.h>

#include <mach/msm_battery.h>

#define BCM_DEBUG 1

static int wifi_status;
static int bt_status;

static struct vreg *vreg_ruim;

void kc_bcm_signal_init(void)
{
	/* Set WiFi REG_ON & RST signal to Low */
	HS_A_REG_ON_LO();
	HS_A_WL_RST_N_LO();

	/* Set WiFi REG_ON & RST signal to Low */
	HS_A_REG_ON_LO();
	HS_A_WL_RST_N_LO();

	/* Initialize WiFi/Bluetooth status */
	/* 0->OFF , 1->ON */
	wifi_status = 0;
	bt_status = 0;

	msm_batt_oem_update_dev_info_wlan( DEV_INFO_WLAN_OFF );
}

void bcm_wlan_power_on(int param)
{
#if BCM_DEBUG
  printk(KERN_INFO "bcm_wlan_power_on() CALLED, param = %d\n", param);
#endif

  HS_A_WL_WAKE_N_LO();

  if (param == 1) /* Turn on wlan power (RUIM, WL_REG_ON, WL_RST_N) */
  {
#if BCM_DEBUG
  	printk(KERN_INFO "Turn on wlan power (RUIM, WL_REG_ON, WL_RST_N) \n");
#endif

	/* Check WiFi status */
	if (wifi_status == 0)
	{
		if (bt_status == 0)
		{
#if BCM_DEBUG
			printk(KERN_INFO "Bluetooth OFF, Turn On Power\n");
#endif
			/* Turn on SDIO Power (ruim) */
			hs_vreg_ruim_enable();

			/* wait 2 clocks */
			msleep(1);
			
			/* Turn on WL_REG_ON */
			HS_A_REG_ON_HI();

		} else {
#if BCM_DEBUG
			printk(KERN_INFO "Bluetooth ON, No need to Turn On Power\n");
#endif
		}

		/* Wait 100ms (Broadcom specication) */
		msleep(100);

		/* Turn on WL_RST */
		HS_A_WL_RST_N_HI();

		/* Wait 1s (temporiraly, for safe) */
		msleep(100);

		/* Change wifi_status OFF -> ON */
		wifi_status = 1;
	}

   } else {  /* Turn on WL_RST_N */ 
#if BCM_DEBUG
  	printk(KERN_INFO "Turn on WL_RST_N (Wake from sleep ? Not deeply consider yet)\n");
#endif
	/* Turn on WL_RST */
	HS_A_WL_RST_N_HI();

	/* Wait 100ms (temporiraly, for safe) */
	msleep(100);	
   }
	msm_batt_oem_update_dev_info_wlan( DEV_INFO_WLAN_ON );
}
EXPORT_SYMBOL(bcm_wlan_power_on);

void bcm_wlan_power_off(int param)
{
#if BCM_DEBUG
  printk(KERN_INFO "bcm_wlan_power_off() CALLED, param = %d\n", param);
#endif

  if (param == 1) /* Turn off wlan power (RUIM, WL_REG_ON, WL_RST_N) */
  {
#if BCM_DEBUG
	printk(KERN_INFO "Turn off wlan power (RUIM, WL_REG_ON, WL_RST_N)\n");
#endif

	/* Check WiFi status */
	if (wifi_status == 1)
	{
		/* Wait 100ms (temporiraly, for safe) */
		msleep(100);

		/* Turn off WL_RST */	
		HS_A_WL_RST_N_LO();

		/* Wait 100ms (temporiraly, for safe) */
		msleep(100);

		if (bt_status == 0)
		{
#if BCM_DEBUG
			printk(KERN_INFO "Bluetooth OFF, Turn Off Power\n");
#endif

			/* Turn off WL_REG_ON */
			HS_A_REG_ON_LO();

			/* Wait 100ms (temporiraly, for safe) */
			msleep(100);

		        /* Turn off SDIO Power (ruim) */
		        hs_vreg_ruim_disable();
		} else {
#if BCM_DEBUG
			printk(KERN_INFO "Bluetooth ON, Cannot Turn Off Power\n");
#endif
		}

		/* Wait 100ms (temporiraly, for safe) */
		//msleep(200);	

		/* Change wifi_status ON -> OFF */
		wifi_status = 0;
	}

   } else { /* Turn off WL_RST_N */ 
#if BCM_DEBUG
  	printk(KERN_INFO "Turn off WL_RST_N (Change to sleep ? Not deeply consider yet)\n");
#endif

	/* Turn off WL_RST */	
	HS_A_WL_RST_N_LO();

	/* Wait 100ms (temporiraly, for safe) */
	msleep(100);	
   }

	msm_batt_oem_update_dev_info_wlan( DEV_INFO_WLAN_OFF );
}
EXPORT_SYMBOL(bcm_wlan_power_off);

/* Bluetooth signal control function */
void kc_bcm_bt_signal_ctl(int on)
{
   if(on)
   { 
	/* Bluetooth Enable */

#if BCM_DEBUG
	printk(KERN_INFO "kc_bcm_bt_signal_ctl() CALLED, ON\n");
#endif

	/* Check bluetooth status */
	if (bt_status == 0)
	{
		if (wifi_status == 0)
		{
#if BCM_DEBUG
			printk(KERN_INFO "WiFi OFF, Turn On Power\n");
#endif
			/* Turn on SDIO Power (ruim) */
			hs_vreg_ruim_enable();

			/* wait 2 clocks */
			msleep(1);

			/* Turn on BT_REG_ON (connected with WL_REG_ON in module)  */
			HS_A_REG_ON_HI();

		} else {
#if BCM_DEBUG
			printk(KERN_INFO "WiFi ON, No need to Turn On Power\n");
#endif
		}

			/* Wait 100ms (Broadcom specication) */
			msleep(100);

			/* Turn off BT_RST */
			HS_A_BT_RST_HI();

			/* Wait 1s (temporiraly, for safe, need to be test) */
			msleep(1000);

			/* Change bt_status OFF -> ON */
			bt_status = 1;
	}

   } else {
	/* Bluetooth Disable */

#if BCM_DEBUG
	printk(KERN_INFO "kc_bcm_bt_signal_ctl() CALLED, OFF\n");
#endif

	/* Check Bluetooth status */
	if (bt_status == 1)
	{
		/* Turn off BT_RST */	
		HS_A_BT_RST_LO();

		if (wifi_status == 0)
		{
#if BCM_DEBUG
			printk(KERN_INFO "WiFi OFF, Turn Off Power\n");
#endif
			/* Turn off BT_REG_ON (connected with WL_REG_ON in module) */
			HS_A_REG_ON_LO();

		        /* Turn off SDIO Power (ruim) */
		        hs_vreg_ruim_disable();

		} else {
#if BCM_DEBUG
			printk(KERN_INFO "WiFi ON, Cannot Turn Off Power\n");
#endif
		}

		/* Change bt_status ON -> OFF */
		bt_status = 0;
	}
   }
}
EXPORT_SYMBOL(kc_bcm_bt_signal_ctl);

/* 2010/01/07 P.Linh Newly implement (not Test yet) <Start> */
void bcm_bt_power_off(void)
{
#if BCM_DEBUG
	printk(KERN_INFO "bcm_bt_power_off() CALLED\n");
#endif

	/* Check Bluetooth status */
	if (bt_status == 1)
	{
		/* Turn off BT_RST */	
		HS_A_BT_RST_LO();

		if (wifi_status == 0)
		{
#if BCM_DEBUG
			printk(KERN_INFO "WiFi OFF, Turn Off Power\n");
#endif
			/* Turn off BT_REG_ON (connected with WL_REG_ON in module) */
			HS_A_REG_ON_LO();

		        /* Turn off SDIO Power (ruim) */
		        hs_vreg_ruim_disable();

		} else {
#if BCM_DEBUG
			printk(KERN_INFO "WiFi ON, Cannot Turn Off Power\n");
#endif
		}

		/* Change bt_status ON -> OFF */
		bt_status = 0;
	}
}
EXPORT_SYMBOL(bcm_bt_power_off);

void bcm_bt_power_on(void)
{
#if BCM_DEBUG
	printk(KERN_INFO "bcm_bt_power_on() CALLED\n");
#endif

	/* Check bluetooth status */
	if (bt_status == 0)
	{
		if (wifi_status == 0)
		{
#if BCM_DEBUG
			printk(KERN_INFO "WiFi OFF, Turn On Power\n");
#endif
			/* Turn on SDIO Power (ruim) */
			hs_vreg_ruim_enable();

			/* wait 2 clocks */
			msleep(1);

			/* Turn on BT_REG_ON (connected with WL_REG_ON in module)  */
			HS_A_REG_ON_HI();

		} else {
#if BCM_DEBUG
			printk(KERN_INFO "WiFi ON, No need to Turn On Power\n");
#endif
		}

			/* Wait 100ms (Broadcom specication) */
			msleep(100);

			/* Turn off BT_RST */
			HS_A_BT_RST_HI();

			/* Wait 1s (temporiraly, for safe, need to be test) */
			msleep(1000);

			/* Change bt_status OFF -> ON */
			bt_status = 1;
	}
}
EXPORT_SYMBOL(bcm_bt_power_on);
/* 2010/01/07 P.Linh Newly implement (not Test yet) <End> */

/* WiFi/Bluetooth (ruim) Power On */
void hs_vreg_ruim_enable( void )
{
  int rc = 0;
  
  /* get vreg_get */
  vreg_ruim = vreg_get(NULL, "ruim");

  rc = vreg_set_level(vreg_ruim, 2850);
  if (!rc)
  {
    rc = vreg_enable(vreg_ruim);
  }

#if BCM_DEBUG
  printk(KERN_INFO "hs_vreg_ruim_enable() results = %d\n", rc);
#endif
}
EXPORT_SYMBOL(hs_vreg_ruim_enable);

/* WiFi/Bluetooth (ruim) Power Off */
void hs_vreg_ruim_disable( void )
{
  int rc = 0;

  /* get vreg_get */
  vreg_ruim = vreg_get(NULL, "ruim");

  rc = vreg_disable(vreg_ruim);

#if BCM_DEBUG
  printk(KERN_INFO "hs_vreg_ruim_disable() results = %d\n", rc);
#endif
}
EXPORT_SYMBOL(hs_vreg_ruim_disable);
