/* This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
 */

#include <linux/platform_device.h>
#include <linux/mutex.h>
#include <linux/syscalls.h> /* sys_sync */
#include <linux/wakelock.h>
#include <linux/module.h>
#include <linux/hrtimer.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/gpio.h>
#include <linux/delay.h>

#include <mach/open_and_close.h>

static struct workqueue_struct *open_and_close_work0;

static int state = -1;

struct open_and_close_local {
	struct work_struct	work0;
};

static struct open_and_close_local open_and_close_data;
static DEFINE_MUTEX(open_and_close_lock);
static LIST_HEAD(open_and_close_handlers);

int get_open_close_state(void)
{
  return state;
}
void register_open_and_close(struct open_and_close *handler)
{
	struct list_head *pos=0;

	mutex_lock(&open_and_close_lock);
	list_for_each(pos, &open_and_close_handlers) {
		struct open_and_close *e;
		e = list_entry(pos, struct open_and_close, link);
		if (e)
			break;
	}
	list_add_tail(&handler->link, pos);
	mutex_unlock(&open_and_close_lock);
}
EXPORT_SYMBOL(register_open_and_close);

void unregister_open_and_close(struct open_and_close *handler)
{
	mutex_lock(&open_and_close_lock);
	list_del(&handler->link);
	mutex_unlock(&open_and_close_lock);
}
EXPORT_SYMBOL(unregister_open_and_close);

void open_and_close_irq(int open)
{
	state = open;
	queue_work(open_and_close_work0, &open_and_close_data.work0);
}
EXPORT_SYMBOL(open_and_close_irq);

static void open_and_close_work_0(struct work_struct *work)
{
	struct open_and_close *pos;
	int openclose = state;

	printk("open_and_close_irq::%d\n",openclose);

	mutex_lock(&open_and_close_lock);

	list_for_each_entry(pos, &open_and_close_handlers, link) {
		if (pos->open_and_close != NULL)
			pos->open_and_close(pos,openclose);
	}
	mutex_unlock(&open_and_close_lock);

}

static int __init open_and_close_init(void)
{
	INIT_WORK(&open_and_close_data.work0, open_and_close_work_0);

	open_and_close_work0 = create_singlethread_workqueue("open_and_close_work0");
}

static void __exit open_and_close_exit(void)
{
	if (open_and_close_work0)
		destroy_workqueue(open_and_close_work0);
}

module_init(open_and_close_init);
module_exit(open_and_close_exit);

MODULE_AUTHOR("Kyocera Corp.");
MODULE_DESCRIPTION("open and close Driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:open_and_close");

