/* 
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
*/

#ifndef _GEO_SENSOR_H_
#define _GEO_SENSOR_H_

#define __LINUX_KERNEL_DRIVER__

#ifndef STATIC
#define STATIC static
#endif

#ifndef NULL
#define NULL ((void*)0)
#endif

#ifndef FALSE
#define FALSE (0)
#endif

#ifndef TRUE
#define TRUE (!(0))
#endif

#ifndef NELEMS
#define NELEMS(a) ((int)(sizeof(a)/sizeof(a[0])))
#endif

#ifdef __LINUX_KERNEL_DRIVER__
#include <linux/types.h>
#include <asm/uaccess.h>
#include <linux/kernel.h>
#define GEO_LOGD(...) printk(__VA_ARGS__)
#define GEO_LOGI(...) printk(__VA_ARGS__)
#define GEO_LOGW(...) printk(KERN_WARNING __VA_ARGS__)
#define GEO_LOGE(...) printk(KERN_ERR __VA_ARGS__)
#else
#include <stdint.h>
#define GEO_LOGD(...) 
#define GEO_LOGI(...) 
#define GEO_LOGW(...) 
#define GEO_LOGE(...) 
#endif

#ifndef __UTIMER_H__
#define __UTIMER_H__

struct utimeval {
    int32_t tv_sec;
    int32_t tv_msec;
};

struct utimer {
    struct utimeval prev_time;
    struct utimeval total_time;
    struct utimeval delay_ms;
};

STATIC int utimeval_init(struct utimeval *val);
STATIC int utimeval_is_initial(struct utimeval *val);
STATIC int utimeval_is_overflow(struct utimeval *val);
STATIC struct utimeval utimeval_plus(struct utimeval *first, struct utimeval *second);
STATIC struct utimeval utimeval_minus(struct utimeval *first, struct utimeval *second);
STATIC int utimeval_greater_than(struct utimeval *first, struct utimeval *second);
STATIC int utimeval_greater_or_equal(struct utimeval *first, struct utimeval *second);
STATIC int utimeval_greater_than_zero(struct utimeval *val);
STATIC int utimeval_less_than_zero(struct utimeval *val);
STATIC struct utimeval *msec_to_utimeval(struct utimeval *result, uint32_t msec);
STATIC uint32_t utimeval_to_msec(struct utimeval *val);

STATIC struct utimeval utimer_calc_next_time(struct utimer *ut,
                                             struct utimeval *cur);
STATIC struct utimeval utimer_current_time(void);
STATIC int utimer_is_timeout(struct utimer *ut);
STATIC int utimer_clear_timeout(struct utimer *ut);
STATIC uint32_t utimer_get_delay(struct utimer *ut);
STATIC int utimer_set_delay(struct utimer *ut, uint32_t delay_ms);
STATIC int utimer_update(struct utimer *ut);
STATIC int utimer_update_with_curtime(struct utimer *ut, struct utimeval *cur);
STATIC uint32_t utimer_sleep_time_with_curtime(struct utimer *ut,
                                               struct utimeval *cur);
STATIC int utimer_init(struct utimer *ut, uint32_t delay_ms);
STATIC int utimer_clear(struct utimer *ut);
STATIC void utimer_lib_init(void (*func)(int *sec, int *msec));

#endif

#ifndef __GEOSENS_CDRIVER_H__
#define __GEOSENS_CDRIVER_H__

# define GEOSENS_CDRV_CENTER_X  512
# define GEOSENS_CDRV_CENTER_Y1 512
# define GEOSENS_CDRV_CENTER_Y2 512
# define GEOSENS_CDRV_CENTER_T  256
# define GEOSENS_CDRV_CENTER_I1 512
# define GEOSENS_CDRV_CENTER_I2 512
# define GEOSENS_CDRV_CENTER_I3 512

#define GEOSENS_CDRV_ROUGHOFFSET_MEASURE_OF_VALUE 33
#define GEOSENS_CDRV_ROUGHOFFSET_MEASURE_UF_VALUE  0
#define GEOSENS_CDRV_NORMAL_MEASURE_OF_VALUE 1024
#define GEOSENS_CDRV_NORMAL_MEASURE_UF_VALUE    1

#define GEOSENS_CMD_MEASURE_ROUGHOFFSET 0x1
#define GEOSENS_CMD_MEASURE_XY1Y2T      0x2

#define GEOSENS_RDSEL_MEASURE     0xe0
#define GEOSENS_RDSEL_CALREGISTER 0xe8

#define GEOSENS_WAIT_MEASURE_ROUGHOFFSET  2 /*  1.5[ms] */
#define GEOSENS_WAIT_MEASURE_XY1Y2T      13 /* 12.3[ms] */

#define GEOSENS_I2C_SLAVE_ADDRESS 0xff
#define GEOSENS_GSENSOR_INITIALIZED     (0x01)
#define GEOSENS_MSENSOR_INITIALIZED     (0x02)

#define GEOSENS_CDRV_NO_ERROR 0
#define GEOSENS_CDRV_ERR_ARG (-1)
#define GEOSENS_CDRV_ERR_NOT_INITIALIZED (-3)
#define GEOSENS_CDRV_ERR_BUSY (-4)
#define GEOSENS_CDRV_ERR_I2CCTRL (-5)
#define GEOSENS_CDRV_ERR_ROUGHOFFSET_NOT_WRITTEN (-126)
#define GEOSENS_CDRV_ERROR (-127)

#define GEOSENS_CDRV_MEASURE_X_OFUF  0x1
#define GEOSENS_CDRV_MEASURE_Y1_OFUF 0x2
#define GEOSENS_CDRV_MEASURE_Y2_OFUF 0x4

struct ami304_machdep_func {
    int (*i2c_open)(void);
    int (*i2c_close)(void);
    int (*i2c_write)(uint8_t slave, const uint8_t *buf, int len);
    int (*i2c_read)(uint8_t slave, uint8_t *buf, int len);
    void (*msleep)(int ms);
};

STATIC int ami304_cdrv_actuate_initcoil(void);
STATIC int ami304_cdrv_set_rough_offset(const uint8_t *rough_offset);
STATIC int ami304_cdrv_recalc_fine_offset(int32_t *prev_fine_offset,
                                          int32_t *new_fine_offset,
                                          uint8_t *prev_rough_offset,
                                          uint8_t *new_rough_offset);
STATIC int ami304_cdrv_set_transformatiom_matrix(const int8_t *transform);
STATIC int ami304_cdrv_measure_rough_offset(uint8_t *rough_offset);
STATIC int ami304_cdrv_measure(int32_t *msens, int32_t *raw, int16_t *t);
STATIC int ami304_cdrv_init(const int8_t *transform,
                            struct ami304_machdep_func *func);
STATIC int ami304_cdrv_term(void);

#endif

#ifndef __GEOSENS_DRIVER_H__
#define __GEOSENS_DRIVER_H__

#define GEOSENS_NO_ERROR                 (0)
#define GEOSENS_ERROR_ARG                (GEOSENS_CDRV_ERR_ARG)
#define GEOSENS_ERROR_NOT_INITIALIZED    (GEOSENS_CDRV_ERR_NOT_INITIALIZED)
#define GEOSENS_ERROR_BUSY               (GEOSENS_CDRV_ERR_BUSY)
#define GEOSENS_ERROR_I2C                (GEOSENS_CDRV_ERR_I2CCTRL)
#define GEOSENS_ERROR_NOT_ACTIVE         (-124)
#define GEOSENS_ERROR_ROUGHOFFSET_NOT_WRITTEN (GEOSENS_CDRV_ERR_ROUGHOFFSET_NOT_WRITTEN)
#define GEOSENS_ERROR_ERROR              (GEOSENS_CDRV_ERROR)
#define GEOSENS_ERROR_RESTARTSYS         (-512)

#define GEOSENS_IOC_GET_DRIVER_STATE     (1)
#define GEOSENS_IOC_SET_DRIVER_STATE     (2)

#define GEOSENS_REPORT_DATA                  (0x01)
#define GEOSENS_REPORT_CALIB                 (0x02)
#define GEOSENS_REPORT_OVERFLOW_OCCURED      (0x04)
#define GEOSENS_REPORT_ROUGH_OFFSET_CHANGED  (0x08)
#define GEOSENS_REPORT_FINE_OFFSET_CHANGED   (0x10)

struct ami304_driver_state {
    int32_t fine_offset[3];
    uint8_t rough_offset[3];
    int accuracy;
};

struct geomagnetic_hwdep_driver {
    int (*init)(void);
    int (*term)(void);
    int (*get_enable)(void);
    int (*set_enable)(int enable);
    int (*get_filter_enable)(void);
    int (*set_filter_enable)(int enable);
    int (*get_filter_len)(void);
    int (*set_filter_len)(int len);
    int (*get_delay)(void);
    int (*set_delay)(int delay);
    int (*get_position)(void);
    int (*set_position)(int accuracy);
    int (*measure)(int32_t *magnetic, int32_t *raw, int32_t *accuracy,
                   uint32_t *time_delay_ms);
    int (*ioctl)(unsigned int cmd, unsigned long args);

    struct geomagnetic_hwdep_callback {
        int (*lock)(void);
        int (*unlock)(void);
        int (*i2c_open)(void);
        int (*i2c_close)(void);
        int (*i2c_write)(uint8_t slave, const uint8_t *buf, int len);
        int (*i2c_read)(uint8_t slave, uint8_t *buf, int len);
        void (*msleep)(int ms);
        void (*current_time)(int32_t *sec, int32_t *msec);
    } callback;
};

STATIC int geomagnetic_driver_init(struct geomagnetic_hwdep_driver *hwdep_driver);

#endif

#ifndef _GEO_SENSOR_LINUX_H_
#define _GEO_SENSOR_LINUX_H_

#define GEOMAGNETIC_I2C_DEVICE_NAME     "geo_sensor"
#define GEOMAGNETIC_DEVICE_NAME         "geo_sensor"
#define GEOMAGNETIC_INPUT_NAME          "geomagnetic"
#define GEOMAGNETIC_INPUT_RAW_NAME      "geomagnetic_raw"

#endif /* _GEO_SENSOR_LINUX_H_ */

#endif /* _GEO_SENSOR_H_ */
