/* This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/leds.h>

#include <mach/pmic.h>
#include <mach/hs_io_ctl_a.h>


extern void dts_touch_key_enable_up(bool onoff);
extern void dts_touch_key_enable_down(bool onoff);

static int msm_btn_bl_suspend = 0;
static int last_value_up = -1;
static int last_value_down = -1;

static void msm_btn_bl_led_set_hw(bool is_up, enum led_brightness value)
{
	enum mpp_which  mpp_1st;
	enum mpp_which  mpp_2nd;
	enum mpp_i_sink_level pm_level;
	enum mpp_i_sink_switch pm_switch;

	if (msm_btn_bl_suspend)
	{
		return;
	}

	if (is_up){
		if (last_value_up == value)
			return;
		
		last_value_up = value;
	
		mpp_1st = PM_MPP_5;
		mpp_2nd = PM_MPP_6;
	}
	else
	{
		if (last_value_down == value)
			return;
		
		last_value_down = value;

		mpp_1st = PM_MPP_9;
		mpp_2nd = PM_MPP_10;
	}

	if (value == 0){
		pm_switch = PM_MPP__I_SINK__SWITCH_DIS;
		pm_level  = PM_MPP__I_SINK__LEVEL_5mA;
	}
	else {
		pm_switch = PM_MPP__I_SINK__SWITCH_ENA;
		pm_level  = PM_MPP__I_SINK__LEVEL_10mA;
	}

	if (HS_INTER_HW_VER >= HS_INTER_HW_VER_WS31){
		pmic_secure_mpp_config_i_sink(mpp_1st, pm_level, pm_switch);
		pmic_secure_mpp_config_i_sink(mpp_2nd, pm_level, PM_MPP__I_SINK__SWITCH_DIS);
	}
	else{
		pm_level  = PM_MPP__I_SINK__LEVEL_5mA;
		pmic_secure_mpp_config_i_sink(mpp_1st, pm_level, pm_switch);
		pmic_secure_mpp_config_i_sink(mpp_2nd, pm_level, pm_switch);
	}
}


static void msm_btn_bl_led_set_up(struct led_classdev *led_cdev,
	enum led_brightness value)
{
	msm_btn_bl_led_set_hw(true, value);
}
static void msm_btn_bl_led_set_down(struct led_classdev *led_cdev,
	enum led_brightness value)
{
	msm_btn_bl_led_set_hw(false, value);
}

static void msm_btn_bl_led_set(struct led_classdev *led_cdev,
	enum led_brightness value)
{
	msm_btn_bl_led_set_up(led_cdev, value);
	msm_btn_bl_led_set_down(led_cdev, value);
}

static struct led_classdev msm_btn_bl_led = {
	.name			= "button-backlight",
	.brightness_set		= msm_btn_bl_led_set,
	.brightness		= LED_OFF,
};

static struct led_classdev msm_btn_bl_led_up = {
	.name			= "button-backlight-up",
	.brightness_set		= msm_btn_bl_led_set_up,
	.brightness		= LED_OFF,
};

static struct led_classdev msm_btn_bl_led_down = {
	.name			= "button-backlight-down",
	.brightness_set		= msm_btn_bl_led_set_down,
	.brightness		= LED_OFF,
};


static int msm_pmic_led_probe(struct platform_device *pdev)
{
	int rc;

	rc = led_classdev_register(&pdev->dev, &msm_btn_bl_led);
	if (rc) {
		dev_err(&pdev->dev, "unable to register led class driver\n");
		return rc;
	}
	rc = led_classdev_register(&pdev->dev, &msm_btn_bl_led_up);
	if (rc) {
		dev_err(&pdev->dev, "unable to register led class driver\n");
		return rc;
	}
	rc = led_classdev_register(&pdev->dev, &msm_btn_bl_led_down);
	if (rc) {
		dev_err(&pdev->dev, "unable to register led class driver\n");
		return rc;
	}
	return rc;
}

static int __devexit msm_pmic_led_remove(struct platform_device *pdev)
{
	led_classdev_unregister(&msm_btn_bl_led);
	led_classdev_unregister(&msm_btn_bl_led_up);
	led_classdev_unregister(&msm_btn_bl_led_down);

	return 0;
}

#ifdef CONFIG_PM
static int msm_pmic_led_suspend(struct platform_device *dev,
		pm_message_t state)
{
	printk("msm_pmic_led_suspend last_up:%d, last_down:%d\n", last_value_up, last_value_down);
	led_classdev_suspend(&msm_btn_bl_led);
	led_classdev_suspend(&msm_btn_bl_led_up);
	led_classdev_suspend(&msm_btn_bl_led_down);

	msm_btn_bl_suspend = 1;

	return 0;
}

static int msm_pmic_led_resume(struct platform_device *dev)
{
	msm_btn_bl_suspend = 0;

	printk("msm_pmic_led_resume last_up:%d, last_down:%d\n", last_value_up, last_value_down);

	led_classdev_resume(&msm_btn_bl_led);
	led_classdev_resume(&msm_btn_bl_led_up);
	led_classdev_resume(&msm_btn_bl_led_down);

	return 0;
}
#else
#define msm_pmic_led_suspend NULL
#define msm_pmic_led_resume NULL
#endif

static struct platform_driver msm_pmic_led_driver = {
	.probe		= msm_pmic_led_probe,
	.remove		= __devexit_p(msm_pmic_led_remove),
	.suspend	= msm_pmic_led_suspend,
	.resume		= msm_pmic_led_resume,
	.driver		= {
		.name	= "pmic-leds-button",
		.owner	= THIS_MODULE,
	},
};

static int __init msm_pmic_led_init(void)
{
	return platform_driver_register(&msm_pmic_led_driver);
}
module_init(msm_pmic_led_init);

static void __exit msm_pmic_led_exit(void)
{
	platform_driver_unregister(&msm_pmic_led_driver);
}
module_exit(msm_pmic_led_exit);

MODULE_DESCRIPTION("MSM PMIC LEDs driver");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("platform:pmic-leds-button");

