/* This software is contributed or developed by KYOCERA Corporation. */
/* (C) 2011 KYOCERA Corporation                                      */


#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/hrtimer.h>

#include <mach/hardware.h>
#include <linux/io.h>

#include <asm/system.h>
#include <asm/mach-types.h>
#include <linux/semaphore.h>

#include <linux/fb.h>

#include "msm_fb.h"
#include "mdp.h"

#define DTS_LCD_DROP_WAIT_TIME  (3)

static struct fb_info          *fbi1 = NULL;
static struct msm_fb_data_type *mfd1 = NULL;
static struct fb_info          *fbi2 = NULL;
static struct msm_fb_data_type *mfd2 = NULL;

#if 0
#define LOCAL_FUNC static
#else
#define LOCAL_FUNC
#endif

LOCAL_FUNC void dts_update_start(struct mdp_dirty_region *dirtyPtr);
LOCAL_FUNC void dts_update_vsync_mddi_prim(void);
LOCAL_FUNC void dts_update_vsync_rgb(void);
LOCAL_FUNC void dts_update_dma_s_done(void);
LOCAL_FUNC void dts_update_dma_p_done(void);
#if DTS_LCD_FBPRINT
LOCAL_FUNC void dts_update_fbprint(void);
#endif

LOCAL_FUNC int set_free_idx(void);
LOCAL_FUNC void dts_udelay(int t);
LOCAL_FUNC int calc_delay(struct msm_fb_data_type *mfd);

#if DTS_LCD_DISPLAY_UPDATE
int is_display_update = 0;
#endif
static boolean mfd1_drop = FALSE;
static boolean mfd2_drop = FALSE;

extern int dts_lcd_check_panel_ready;
extern boolean dts_lcd_flag_under_panel_close;

void dts_update_proc(
    DTS_UPDATE_EVENT  evt,
    struct fb_info   *info,
    void             *data
) {
    if (fbi1 == NULL) {
        if (info == NULL) {
            printk(KERN_ERR "dts_update_proc(DTS_UPDATE_START): info = NULL !\n");
            return;
        }

        fbi1 = info;
        mfd1 = (struct msm_fb_data_type *)fbi1->par;
        fbi2 = mfd1->other_fbi;
        mfd2 = (struct msm_fb_data_type *)fbi2->par;
    }

    switch(evt) {
    case DTS_UPDATE_START:
        dts_update_start((struct mdp_dirty_region *)data);
        break;
    case DTS_UPDATE_VSYNC_MDDI_PRIM:
        dts_update_vsync_mddi_prim();
        break;
    case DTS_UPDATE_VSYNC_RGB:
        dts_update_vsync_rgb();
        break;
    case DTS_UPDATE_DMA_S_DONE:
        dts_update_dma_s_done();
        break;
    case DTS_UPDATE_DMA_P_DONE:
        dts_update_dma_p_done();
        break;
#if DTS_LCD_FBPRINT
    case DTS_UPDATE_FBPRINT:
        dts_update_fbprint();
        break;
#endif
    default:
        break;
    }
}

void dts_clear_ibuf2(void)
{
    int i;
    int bpp;
    int size;

    if (mfd2 == NULL || fbi2 == NULL) {
        return;
    }

    bpp  = 2;
    size = mfd2->var_xres * mfd2->var_yres * 2 / 2;

    for(i=0; i<DTS_LCD_NUMBER_OF_IBUF; i++){
        if (mfd2->ibuf_tbl[i].buf != NULL) {
            memset(DTS_SMI_TO_VIRT(mfd2->ibuf_tbl[i].buf, fbi2), 0x0, size);
        }
    }
}

LOCAL_FUNC void dts_update_start(
    struct mdp_dirty_region *dirtyPtr
) {
    boolean updateFlg1 = TRUE;
	struct mdp_dirty_region *dirtyPtr1 = NULL;
    boolean updateFlg2 = TRUE;
	struct mdp_dirty_region *dirtyPtr2 = NULL;
	struct mdp_dirty_region dirty1;
	struct mdp_dirty_region dirty2;

    if (dirtyPtr != NULL) {
        int x_center = fbi1->var.xres / DTS_LCD_SCALE;
        dirty1 = *dirtyPtr;
        dirty2 = *dirtyPtr;

        if (dirtyPtr->xoffset >= x_center) {
            dirty1.xoffset -= x_center;
            dirtyPtr1 = &dirty1;
            updateFlg2 = FALSE;
        } else if (dirtyPtr->xoffset + dirtyPtr->width <= x_center) {
            dirtyPtr2 = &dirty2;
            updateFlg1 = FALSE;
        } else {
            dirty1.xoffset = 0;
            dirty1.width   = dirtyPtr->xoffset + dirtyPtr->width - x_center;
            dirty2.width   = x_center - dirtyPtr->xoffset;

            dirtyPtr1 = &dirty1;
            dirtyPtr2 = &dirty2;
        }
    }

    if (mfd1->ibuf_idx_ppp_vsync >= 0 && mfd1_drop == TRUE) {
        wait_for_completion_timeout(&mfd1->ibuf_idx_ppp_comp, DTS_LCD_DROP_WAIT_TIME);
    }
    if (mfd2->ibuf_idx_ppp_vsync >= 0 && mfd2_drop == TRUE && dts_lcd_flag_under_panel_close == FALSE) {
        wait_for_completion_timeout(&mfd2->ibuf_idx_ppp_comp, DTS_LCD_DROP_WAIT_TIME);
    }

    if (mfd2->ibuf_idx_ppp_vsync >= 0) {
        mfd2->ibuf_idx_next = mfd2->ibuf_idx_ppp_vsync;
        mfd2->dma->waiting = FALSE;
        dirtyPtr2 = NULL;
        mfd2->ibuf_idx_ppp_vsync = -1;
        mfd2_drop = TRUE;
    }
    else
    {
      mfd2_drop = FALSE;
    }

    if (updateFlg2 && dts_lcd_flag_under_panel_close == FALSE)
    {
        mdp_set_dma_pan_info(fbi2, dirtyPtr2, TRUE);

        if(dts_lcd_check_panel_ready)
        {
            mdp_ppp_ibuf(fbi2);

            mdp_dma_pan_update(fbi2, FALSE);
        }
    }


    if (mfd1->ibuf_idx_ppp_vsync >= 0) {
        mfd1->ibuf_idx_next = mfd1->ibuf_idx_ppp_vsync;
        mfd1->dma->waiting = FALSE;
        dirtyPtr1 = NULL;
        mfd1->ibuf_idx_ppp_vsync = -1;
        mfd1_drop = TRUE;
    }
    else
    {
      mfd1_drop = FALSE;
    }

#if DTS_LCD_FBPRINT
    mdp_ppp_fbprint(fbi1);
    if (dirtyPtr != NULL) {
        dirtyPtr1 = &dirty1;
    }
#endif

    if (updateFlg1) {
        mdp_set_dma_pan_info(fbi1, dirtyPtr1, TRUE);
        
        if(dts_lcd_check_panel_ready)
        {
            mdp_ppp_ibuf(fbi1);

            mdp_dma_pan_update(fbi1, FALSE);
        }
    }

}

LOCAL_FUNC void dts_update_vsync_mddi_prim(void)
{
    if (mfd1->dma->busy)
        return;

    mfd1->dma->waiting = FALSE;
    vsync_disable_irq_nosync(UPDATE_VSYNC_MDDI);

    dts_udelay(calc_delay(mfd1));

    mdp_dma_pan_update_main(fbi1);
}

LOCAL_FUNC void dts_update_vsync_rgb(void)
{
    if (mfd2->dma->busy)
        return;

    mfd2->dma->waiting = FALSE;
    vsync_disable_irq_nosync(UPDATE_VSYNC_RGB);

    mdp_dma_pan_update_main(fbi2);
}

LOCAL_FUNC void dts_update_dma_s_done(void)
{
    mdp_dma_pan_update_finish(fbi1);
}

LOCAL_FUNC void dts_update_dma_p_done(void)
{
    mdp_dma_pan_update_finish(fbi2);
}

/*-------------------------------------------------------------------*/
#if DTS_LCD_FBPRINT
LOCAL_FUNC void dts_update_fbprint(void)
{
    if (mfd1->ibuf_idx_ppp_vsync >= 0 ||
        mfd1->ibuf_idx_dma       >= 0) {
        return;
    } else {
        mfd1->ibuf_idx_dma = DTS_LCD_NUMBER_OF_IBUF;
    }

    mdp_dma_fbprint(fbi1);
}
#endif

LOCAL_FUNC void dts_udelay(
    int t
) {
    while (t >= MAX_UDELAY_MS * 1000) {
        udelay(MAX_UDELAY_MS * 1000);
        t -= MAX_UDELAY_MS * 1000;
    }
    if (t > 0) {
        udelay(t);
    }
}

LOCAL_FUNC int calc_delay(
    struct msm_fb_data_type *mfd
) {
    MDPIBUF *iBuf = &mfd->ibuf_tbl[mfd->ibuf_idx_ppp_vsync];
    int w = mfd->var_xres / DTS_LCD_SCALE;
    int h = mfd->var_yres;

    return (16 * 1000 *                        
            ((w - iBuf->dma_w) * iBuf->dma_h + 
             iBuf->dma_y * w) /                
            (w * h) +                          
            195);                              
}

#endif 
